"""
Forms definition
"""

# Django
from django import forms
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.forms import ModelForm
from django.utils.translation import gettext_lazy as _

# CKEditor
from django_ckeditor_5.widgets import CKEditor5Widget

# AA Bulletin Board
from aa_bulletin_board.helper.string import string_cleanup
from aa_bulletin_board.models import Bulletin


class SpecialModelChoiceIterator(forms.models.ModelChoiceIterator):
    """
    Variant of Django's ModelChoiceIterator to prevent it from always re-fetching the
    given queryset from the database.
    """

    def __iter__(self):
        if self.field.empty_label is not None:
            yield "", self.field.empty_label

        queryset = self.queryset

        for obj in queryset:
            yield self.choice(obj=obj)


class SpecialModelMultipleChoiceField(forms.ModelMultipleChoiceField):
    """
    Variant of Django's ModelMultipleChoiceField to prevent it from always
    re-fetching the given queryset from the database.
    """

    iterator = SpecialModelChoiceIterator

    def _get_queryset(self):
        return self._queryset

    def _set_queryset(self, queryset):
        self._queryset = queryset
        self.widget.choices = self.choices

    queryset = property(fget=_get_queryset, fset=_set_queryset)


class BulletinForm(ModelForm):
    """
    Form for bulletins
    """

    groups = SpecialModelMultipleChoiceField(
        required=False,
        queryset=Group.objects.all(),
        help_text=_(
            "Restrict this bulletin to certain groups. "
            "If no group restrictions are in place, everyone who has access to "
            "this module can read this bulletin."
        ),
    )

    def __init__(self, *args, **kwargs):
        groups_queryset = kwargs.pop("groups_queryset", None)

        super().__init__(*args, **kwargs)

        if groups_queryset:
            self.fields["groups"].queryset = groups_queryset

        # We have to set this to False, otherwise CKEditor5 will not work
        self.fields["content"].required = False

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Form Meta
        """

        model = Bulletin
        fields = ["title", "content", "groups"]
        widgets = {
            "content": CKEditor5Widget(
                config_name="extends",
                attrs={
                    "class": "aa-bulletin-board-ckeditor django_ckeditor_5",
                    "rows": 10,
                    "cols": 20,
                    "style": "width: 100%;",
                },
            )
        }

    def clean(self):
        """
        Clean the form

        :return:
        :rtype:
        """

        cleaned_data = super().clean()
        content = cleaned_data.get("content")

        if not content or content.strip() == "":
            self.add_error(
                "content", ValidationError(_("You have forgotten the content!"))
            )

        cleaned_data["content"] = content.strip()

        return cleaned_data

    def clean_content(self) -> str:
        """
        Cleanup the content

        :return:
        :rtype:
        """

        content = string_cleanup(string=self.cleaned_data["content"])

        return content



"""
App config
"""

# Django
from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

# AA Bulletin Board
from aa_bulletin_board import __version__
from aa_bulletin_board.constants import PACKAGE_NAME


class AaBulletinBoardConfig(AppConfig):
    """
    Application config
    """

    name = PACKAGE_NAME
    label = PACKAGE_NAME
    # Translators: This is the app name and version, which will appear in the Django Backend
    verbose_name = _(f"Bulletin Board v{__version__}")



"""
Test for the UI
"""

# Third Party
from django_webtest import WebTest
from faker import Faker

# Django
from django.contrib.auth.models import Group
from django.urls import reverse

# AA Bulletin Board
from aa_bulletin_board.helper.string import string_cleanup
from aa_bulletin_board.models import Bulletin
from aa_bulletin_board.tests.utils import create_fake_user

fake = Faker()


class TestBulletinUI(WebTest):
    """
    Test Bulletin UI
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")

        # User cannot access bulletins
        cls.user_1001 = create_fake_user(
            character_id=1001, character_name="Peter Parker"
        )

        # User can access bulletins
        cls.user_1002 = create_fake_user(
            character_id=1002,
            character_name="Bruce Wayne",
            permissions=["aa_bulletin_board.basic_access"],
        )

        # User can manage bulletins
        cls.user_1003 = create_fake_user(
            character_id=1003,
            character_name="Clark Kent",
            permissions=[
                "aa_bulletin_board.basic_access",
                "aa_bulletin_board.manage_bulletins",
            ],
        )

    def test_should_show_bulletin_page(self) -> None:
        """
        Test if a bulletin is shown

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Test Bulletin",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1002,
        )
        self.app.set_user(user=self.user_1002)

        # when
        page = self.app.get(
            url=reverse(
                viewname="aa_bulletin_board:view_bulletin", args=[bulletin.slug]
            )
        )

        # then
        self.assertTemplateUsed(
            response=page,
            template_name="aa_bulletin_board/bulletin.html",
        )

    def test_should_redirect_to_bulletin_dashboard_when_bulletin_does_not_exist(
        self,
    ) -> None:
        """
        Test should redirect to bulletin dashboard when bulletin does not exist

        :return:
        :rtype:
        """

        # given
        self.app.set_user(user=self.user_1002)

        # when
        page = self.app.get(
            url=reverse(viewname="aa_bulletin_board:view_bulletin", args=["foobar"])
        )

        # then
        self.assertRedirects(response=page, expected_url="/bulletin-board/")

    def test_should_show_create_bulletin_page(self) -> None:
        """
        Test if create bulletin is shown

        :return:
        :rtype:
        """

        # given
        self.app.set_user(user=self.user_1003)

        # when
        page = self.app.get(url=reverse(viewname="aa_bulletin_board:create_bulletin"))

        # then
        self.assertTemplateUsed(
            response=page,
            template_name="aa_bulletin_board/edit-bulletin.html",
        )

    def test_should_show_edit_bulletin_page(self) -> None:
        """
        Test if edit bulletin is shown

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Test Bulletin",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1002,
        )
        self.app.set_user(user=self.user_1003)

        # when
        page = self.app.get(
            url=reverse(
                viewname="aa_bulletin_board:edit_bulletin", args=[bulletin.slug]
            )
        )

        # then
        self.assertTemplateUsed(
            response=page,
            template_name="aa_bulletin_board/edit-bulletin.html",
        )

    def test_should_redirect_to_bulletin_dashboard_when_edit_bulletin_does_not_exist(
        self,
    ) -> None:
        """
        Test should redirect to bulletin dashboard when bulletin to edit does not exist

        :return:
        :rtype:
        """

        # given
        self.app.set_user(user=self.user_1003)

        # when
        page = self.app.get(
            url=reverse(viewname="aa_bulletin_board:edit_bulletin", args=["foobar"])
        )

        # then
        self.assertRedirects(response=page, expected_url="/bulletin-board/")

    def test_should_return_cleaned_message_string_on_bulletin_creation(self) -> None:
        """
        Test should return a clean/sanitized message string when a new bulletin is created

        :return:
        :rtype:
        """

        # given
        self.app.set_user(user=self.user_1003)
        page = self.app.get(url=reverse(viewname="aa_bulletin_board:create_bulletin"))
        dirty_message = (
            'this is a script test. <script type="text/javascript">alert('
            "'test')</script>and this is style test. <style>.MathJax, "
            ".MathJax_Message, .MathJax_Preview{display: none}</style>end tests."
        )
        cleaned_message = string_cleanup(string=dirty_message)

        # when
        form = page.forms["aa-bulletin-board-bulletin-form"]
        form["title"] = "Message Cleanup Test"
        form["content"] = dirty_message
        form.submit().follow()

        # then
        new_bulletin = Bulletin.objects.last()
        self.assertEqual(first=new_bulletin.content, second=cleaned_message)

    def test_should_return_cleaned_message_string_on_bulletin_edit(self) -> None:
        """
        Test should return a clean/sanitized message string when a bulletin is edited

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Test Bulletin 2",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1002,
        )
        self.app.set_user(user=self.user_1003)

        # when
        page = self.app.get(
            url=reverse(
                viewname="aa_bulletin_board:edit_bulletin", args=[bulletin.slug]
            )
        )
        dirty_message = (
            'this is a script test. <script type="text/javascript">alert('
            "'test')</script>and this is style test. <style>.MathJax, "
            ".MathJax_Message, .MathJax_Preview{display: none}</style>end tests."
        )
        cleaned_message = string_cleanup(string=dirty_message)

        form = page.forms["aa-bulletin-board-bulletin-form"]
        form["title"] = "Message Cleanup Test"
        form["content"] = dirty_message
        form.submit().follow()

        # then
        bulletin_edited = Bulletin.objects.get(pk=bulletin.pk)
        self.assertEqual(first=bulletin_edited.content, second=cleaned_message)

    def test_should_return_to_edit_bulletin_form_for_invalid_form_submitted_on_create_bulletin(
        self,
    ) -> None:
        """
        Test should return to the bulletin form, because the submitted form is not valid
        due to missing title

        :return:
        :rtype:
        """

        self.app.set_user(user=self.user_1003)
        page = self.app.get(url=reverse(viewname="aa_bulletin_board:create_bulletin"))

        # when
        form = page.forms["aa-bulletin-board-bulletin-form"]
        form["content"] = "Lorem Ipsum"
        page = form.submit()

        # then
        self.assertTemplateUsed(
            response=page,
            template_name="aa_bulletin_board/edit-bulletin.html",
        )

    def test_should_return_to_edit_bulletin_form_for_invalid_form_submitted_on_edit_bulletin(
        self,
    ) -> None:
        """
        Test should return to the bulletin form, because the submitted form is not valid
        due to missing title

        :return:
        :rtype:
        """

        bulletin = Bulletin.objects.create(
            title="Test Bulletin",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1002,
        )
        self.app.set_user(user=self.user_1003)

        # when
        page = self.app.get(
            url=reverse(
                viewname="aa_bulletin_board:edit_bulletin", args=[bulletin.slug]
            )
        )

        # when
        form = page.forms["aa-bulletin-board-bulletin-form"]
        form["title"] = ""
        form["content"] = "Lorem Ipsum"
        page = form.submit()

        # then
        self.assertTemplateUsed(
            response=page,
            template_name="aa_bulletin_board/edit-bulletin.html",
        )



"""
Test for aa_bulletin_board.forms
"""

# Django
from django.contrib.auth.models import Group
from django.test import TestCase

# AA Bulletin Board
from aa_bulletin_board.forms import (
    BulletinForm,
    SpecialModelChoiceIterator,
    SpecialModelMultipleChoiceField,
)


class TestSpecialModelChoiceIterator(TestCase):
    """
    Tests for SpecialModelChoiceIterator
    """

    def test_iterates_over_queryset(self):
        """
        Test iterates over the queryset without re-fetching it

        :return:
        :rtype:
        """

        group1 = Group.objects.create(name="Group 1")
        group2 = Group.objects.create(name="Group 2")

        field = SpecialModelMultipleChoiceField(queryset=Group.objects.all())
        iterator = SpecialModelChoiceIterator(field)

        choices = list(iterator)

        self.assertIn((group1.pk, str(group1)), choices)
        self.assertIn((group2.pk, str(group2)), choices)

    def test_includes_empty_label_if_set(self):
        """
        Test includes empty label if a label is set

        :return:
        :rtype:
        """

        field = SpecialModelMultipleChoiceField(queryset=Group.objects.all())
        field.empty_label = "Select a group"
        iterator = SpecialModelChoiceIterator(field)

        choices = list(iterator)

        self.assertIn(("", "Select a group"), choices)


class TestSpecialModelMultipleChoiceField(TestCase):
    """
    Tests for SpecialModelMultipleChoiceField
    """

    def test_sets_and_gets_queryset(self):
        """
        Test sets and gets the queryset without re-fetching it

        :return:
        :rtype:
        """

        field = SpecialModelMultipleChoiceField(queryset=Group.objects.none())
        group1 = Group.objects.create(name="Group 1")
        group2 = Group.objects.create(name="Group 2")
        field.queryset = Group.objects.all()

        self.assertEqual(list(field.queryset), [group1, group2])

    def test_updates_widget_choices_when_queryset_is_set(self):
        """
        Test updates the widget choices when the queryset is set

        :return:
        :rtype:
        """

        field = SpecialModelMultipleChoiceField(queryset=Group.objects.none())
        group1 = Group.objects.create(name="Group 1")
        group2 = Group.objects.create(name="Group 2")
        field.queryset = Group.objects.all()

        self.assertIn((group1.pk, str(group1)), field.widget.choices)
        self.assertIn((group2.pk, str(group2)), field.widget.choices)

    def test_handles_empty_queryset(self):
        """
        Test handles empty queryset

        :return:
        :rtype:
        """

        field = SpecialModelMultipleChoiceField(queryset=Group.objects.none())

        self.assertEqual(list(field.queryset), [])
        self.assertEqual(list(field.widget.choices), [])


class TestBulletinForm(TestCase):
    """
    Tests for BulletinForm
    """

    def test_valid_form_with_groups_queryset(self):
        """
        Test valid form with groups queryset

        :return:
        :rtype:
        """

        group1 = Group.objects.create(name="Group 1")
        group2 = Group.objects.create(name="Group 2")
        form_data = {
            "title": "Test Bulletin",
            "content": "This is a test bulletin.",
            "groups": [group1.pk, group2.pk],
        }
        form = BulletinForm(data=form_data, groups_queryset=Group.objects.all())

        self.assertTrue(form.is_valid())

    def test_invalid_form_without_title(self):
        """
        Test invalid form without title

        :return:
        :rtype:
        """

        form_data = {
            "title": "",
            "content": "This is a test bulletin.",
        }
        form = BulletinForm(data=form_data)

        self.assertFalse(form.is_valid())
        self.assertIn("title", form.errors)

    def test_clean_raises_validation_error_for_empty_content(self):
        """
        Test clean raises validation error for empty content

        :return:
        :rtype:
        """

        form_data = {
            "title": "Test Bulletin",
            "content": "   ",
        }
        form = BulletinForm(data=form_data)

        self.assertFalse(form.is_valid())
        self.assertIn("content", form.errors)

    def test_clean_content_strips_unwanted_characters(self):
        """
        Test clean content strips unwanted characters

        :return:
        :rtype:
        """

        form_data = {
            "title": "Test Bulletin",
            "content": "  <p>Test content</p>  ",
        }
        form = BulletinForm(data=form_data)
        form.is_valid()

        self.assertEqual(form.cleaned_data["content"], "<p>Test content</p>")



# Generated by Django 4.0.10 on 2023-09-22 17:08

# Django
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("aa_bulletin_board", "0004_model_translation"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    ("basic_access", "Can access this app"),
                    ("manage_bulletins", "Can manage (add/change/remove) bulletins"),
                ),
                "verbose_name": "Bulletin Board",
            },
        ),
    ]



# Generated by Django 4.2.10 on 2024-02-11 09:40

# Django
from django.db import migrations

# CKEditor
import django_ckeditor_5.fields


class Migration(migrations.Migration):

    dependencies = [
        ("aa_bulletin_board", "0005_alter_general_options"),
    ]

    operations = [
        migrations.AlterField(
            model_name="bulletin",
            name="content",
            field=django_ckeditor_5.fields.CKEditor5Field(
                default=None, verbose_name="Content"
            ),
        ),
    ]



"""
Testing access to bulletins
"""

# Standard Library
from http import HTTPStatus

# Third Party
from faker import Faker

# Django
from django.contrib.auth.models import Group
from django.test import TestCase
from django.urls import reverse

# AA Bulletin Board
from aa_bulletin_board.models import Bulletin
from aa_bulletin_board.tests.utils import create_fake_user

fake = Faker()


class TestAccess(TestCase):
    """
    Test access to the module
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")

        # User cannot access bulletins
        cls.user_1001 = create_fake_user(
            character_id=1001, character_name="Peter Parker"
        )

        # User can access bulletins
        cls.user_1002 = create_fake_user(
            character_id=1002,
            character_name="Bruce Wayne",
            permissions=["aa_bulletin_board.basic_access"],
        )

        # User can manage bulletins
        cls.user_1003 = create_fake_user(
            character_id=1003,
            character_name="Clark Kent",
            permissions=[
                "aa_bulletin_board.basic_access",
                "aa_bulletin_board.manage_bulletins",
            ],
        )

    def test_should_show_dashboard(self) -> None:
        """
        Test that a user with basic_access can see the bulletin board

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1002)

        # when
        res = self.client.get(path=reverse(viewname="aa_bulletin_board:dashboard"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_bulletin(self) -> None:
        """
        Test that a user with basic_access can see the bulletin

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        self.client.force_login(user=self.user_1002)

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_bulletin_board:view_bulletin",
                kwargs={"slug": bulletin.slug},
            )
        )
        result = Bulletin.objects.user_has_access(user=self.user_1002)

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertIn(member=bulletin, container=result)

    def test_should_not_show_dashboard(self) -> None:
        """
        Test that a user without basic_access can't see the bulletin board

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(path=reverse(viewname="aa_bulletin_board:dashboard"))

        # then
        self.assertIsNot(expr1=res.status_code, expr2=HTTPStatus.OK)

    def test_should_not_show_bulletin(self) -> None:
        """
        Test that a user without basic_access can't see the bulletin

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_bulletin_board:view_bulletin",
                kwargs={"slug": bulletin.slug},
            )
        )

        # then
        self.assertIsNot(expr1=res.status_code, expr2=HTTPStatus.OK)

    def test_should_return_bulletin_for_user_with_perm_manage_bulletins(self) -> None:
        """
        Test that a user with "aa_bulletin_board.manage_bulletins" can see all bulletins

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        # when
        result = Bulletin.objects.user_has_access(user=self.user_1003)

        # then
        self.assertIn(member=bulletin, container=result)

    def test_should_return_restricted_bulletin_for_user_with_perm_manage_bulletins(
        self,
    ) -> None:
        """
        Test that a user with "aa_bulletin_board.manage_bulletins"
        can see all bulletins, even restricted

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        bulletin.groups.add(self.group)

        # when
        result = Bulletin.objects.user_has_access(user=self.user_1003)

        # then
        self.assertIn(member=bulletin, container=result)

    def test_should_return_bulletin_with_no_groups(self) -> None:
        """
        Test that any user with access can se non-restricted bulletins

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        # when
        result = Bulletin.objects.user_has_access(user=self.user_1002)

        # then
        self.assertIn(member=bulletin, container=result)

    def test_should_return_bulletin_for_group_member(self) -> None:
        """
        Test that group restricted bulletins are visible for users who have this group

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        bulletin.groups.add(self.group)
        self.user_1001.groups.add(self.group)

        # when
        result = Bulletin.objects.user_has_access(user=self.user_1001)

        # then
        self.assertIn(member=bulletin, container=result)

    def test_should_not_return_bulletin_for_non_group_member(self) -> None:
        """
        Test that group restricted bulletins are not visible for
        users who don't have this group

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title="Physics",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1002,
        )
        bulletin.groups.add(self.group)

        # when
        result = Bulletin.objects.user_has_access(user=self.user_1002)

        # then
        self.assertNotIn(member=bulletin, container=result)



"""
Hook into AA
"""

# Alliance Auth
from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

# AA Bulletin Board
from aa_bulletin_board import __title__, urls


class AaBulletinBoardMenuItem(MenuItemHook):  # pylint: disable=too-few-public-methods
    """
    This class ensures only authorized users will see the menu entry
    """

    def __init__(self):
        """
        Setup menu entry for sidebar
        """

        MenuItemHook.__init__(
            self,
            text=__title__,
            classes="fa-solid fa-clipboard-list",
            url_name="aa_bulletin_board:dashboard",
            navactive=["aa_bulletin_board:"],
        )

    def render(self, request):
        """
        Check if the user has the permission to view this app

        :param request:
        :type request:
        :return:
        :rtype:
        """

        return (
            MenuItemHook.render(self, request=request)
            if request.user.has_perm("aa_bulletin_board.basic_access")
            else ""
        )


@hooks.register("menu_item_hook")
def register_menu():
    """
    Register our menu item

    :return:
    :rtype:
    """

    return AaBulletinBoardMenuItem()


@hooks.register("url_hook")
def register_urls():
    """
    Register our base url

    :return:
    :rtype:
    """

    return UrlHook(
        urls=urls, namespace="aa_bulletin_board", base_url=r"^bulletin-board/"
    )



"""
Managers for our models
"""

# Django
from django.contrib.auth.models import User
from django.db import models
from django.db.models import Q


class BulletinQuerySet(models.QuerySet):
    """
    Bulletin queryset
    """

    def user_has_access(self, user: User) -> models.QuerySet:
        """
        Filter boards that given user has access to.

        :param user:
        :type user:
        :return:
        :rtype:
        """

        # Forum managers always have access, so assign this permission wisely
        if user.has_perm(perm="aa_bulletin_board.manage_bulletins"):
            return self

        # If not a forum manager, check if the user has access to the board
        return self.filter(
            Q(groups__in=user.groups.all()) | Q(groups__isnull=True)
        ).distinct()


class BulletinManagerBase(models.Manager):  # pylint: disable=too-few-public-methods
    """
    BoardManagerBase
    """

    pass  # pylint: disable=unnecessary-pass


BulletinManager = BulletinManagerBase.from_queryset(queryset_class=BulletinQuerySet)



"""
Our little helpers
"""

# Standard Library
import re


def string_cleanup(string: str) -> str:
    """
    Clean up a string by removing JS, CSS and Head tags

    :param string:
    :type string:
    :return:
    :rtype:
    """

    # Strip unwanted content (JS, CSS, HTML-Head)
    return re.compile(
        pattern=r"<\s*(head|script|style)[^>]*>.*?<\s*/\s*\1\s*>", flags=re.S | re.I
    ).sub(repl="", string=string)



"""
Constants used in this app
"""

# Standard Library
import os

# Django
from django.utils.translation import gettext_lazy as _

APP_NAME = "aa-bulletin-board"
APP_NAME_VERBOSE = "AA Bulletin Board"
APP_TITLE = _("Bulletin Board")
PACKAGE_NAME = "aa_bulletin_board"
APP_BASE_DIR = os.path.join(os.path.dirname(__file__))
APP_STATIC_DIR = os.path.join(APP_BASE_DIR, "static", PACKAGE_NAME)
GITHUB_URL = f"https://github.com/ppfeufer/{APP_NAME}"



"""
The models
"""

# Standard Library
from typing import ClassVar

# Third Party
import unidecode

# Django
from django.contrib.auth.models import Group, User
from django.db import models, transaction
from django.utils.text import slugify
from django.utils.translation import gettext as _

# CKEditor
from django_ckeditor_5.fields import CKEditor5Field

# AA Bulletin Board
from aa_bulletin_board.constants import APP_TITLE
from aa_bulletin_board.helper.string import string_cleanup
from aa_bulletin_board.managers import BulletinManager


def get_sentinel_user() -> User:
    """
    Get the sentinel user or create one

    :return:
    :rtype:
    """

    return User.objects.get_or_create(username="deleted")[0]


def get_bulletin_slug_from_title(bulletin_title: str) -> str:
    """
    Get the slug from the title

    :param bulletin_title:
    :type bulletin_title:
    :return:
    :rtype:
    """

    base_slug = slugify(value=unidecode.unidecode(bulletin_title), allow_unicode=True)
    bulletin_slug = base_slug
    run = 0

    while Bulletin.objects.filter(slug=bulletin_slug).exists():
        run += 1
        bulletin_slug = f"{base_slug}-{run}"

    return bulletin_slug


class General(models.Model):
    """
    Meta model for app permissions
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        verbose_name = APP_TITLE
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", _("Can access this app")),
            ("manage_bulletins", _("Can manage (add/change/remove) bulletins")),
        )


class Bulletin(models.Model):
    """
    Bulletin model
    """

    title = models.CharField(max_length=255, verbose_name=_("Title"))
    slug = models.SlugField(max_length=255, unique=True, allow_unicode=True)
    content = CKEditor5Field(
        blank=False, default=None, verbose_name=_("Content"), config_name="extends"
    )
    created_date = models.DateTimeField(
        auto_now_add=True,
        db_index=True,
        # Translators: This is the date and time the bulletin has been created
        verbose_name=_("Created"),
    )
    updated_date = models.DateTimeField(
        auto_now=True,
        null=True,
        # Translators: This is the date and time the bulletin has been updated
        verbose_name=_("Updated"),
    )
    created_by = models.ForeignKey(
        to=User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(get_sentinel_user),
        verbose_name=_("User"),
    )
    groups = models.ManyToManyField(
        to=Group,
        blank=True,
        related_name="aa_bulletin_board_group_restriction",
        verbose_name=_("Group restrictions"),
    )

    objects: ClassVar[BulletinManager] = BulletinManager()

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Bulletin")
        verbose_name_plural = _("Bulletins")

    def __str__(self) -> str:
        return str(self.title)

    @transaction.atomic()
    def save(self, *args, **kwargs) -> None:
        """
        Add the slug on save

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        self.content = string_cleanup(string=self.content)

        if not self.slug:
            self.slug = get_bulletin_slug_from_title(bulletin_title=self.title)

        super().save(*args, **kwargs)



"""
The views
"""

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.models import Group
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import Prefetch
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect, render
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Bulletin Board
from aa_bulletin_board import __title__
from aa_bulletin_board.forms import Bulletin, BulletinForm

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


@login_required
@permission_required(perm="aa_bulletin_board.basic_access")
def dashboard(request: WSGIRequest) -> HttpResponse:
    """
    Index view

    :param request:
    :type request:
    :return:
    :rtype:
    """

    bulletins = (
        Bulletin.objects.prefetch_related(
            Prefetch(lookup="groups", queryset=Group.objects.order_by("name"))
        )
        .user_has_access(user=request.user)
        .order_by("-created_date")
    )

    context = {"bulletins": bulletins}

    return render(
        request=request,
        template_name="aa_bulletin_board/dashboard.html",
        context=context,
    )


@login_required
@permission_required(perm="aa_bulletin_board.manage_bulletins")
def create_bulletin(request: WSGIRequest) -> HttpResponse:
    """
    Edit an existing bulletin

    :param request:
    :type request:
    :return:
    :rtype:
    """

    # Create an empty form
    form = BulletinForm()

    # If we have a POST request
    if request.method == "POST":
        # Create a form instance from the POST request and populate it with data
        form = BulletinForm(data=request.POST)

        # Check whether it's valid
        if form.is_valid():
            bulletin__title = form.cleaned_data["title"]
            bulletin__content = form.cleaned_data["content"]
            bulletin__created_date = timezone.now()

            # Save the new bulletin
            bulletin = Bulletin(
                title=bulletin__title,
                content=bulletin__content,
                created_date=bulletin__created_date,
                created_by=request.user,
            )
            bulletin.save()

            bulletin.groups.set(  # pylint: disable=no-member
                form.cleaned_data["groups"]
            )

            messages.success(
                request=request, message=_(f'Bulletin "{bulletin__title}" created.')
            )

            return redirect(to="aa_bulletin_board:view_bulletin", slug=bulletin.slug)

    context = {"form": form, "bulletin": False}

    return render(
        request=request,
        template_name="aa_bulletin_board/edit-bulletin.html",
        context=context,
    )


@login_required
@permission_required(perm="aa_bulletin_board.basic_access")
def view_bulletin(request: WSGIRequest, slug: str) -> HttpResponse:
    """
    View a bulletin

    :param request:
    :type request:
    :param slug:
    :type slug:
    :return:
    :rtype:
    """

    try:
        bulletin: Bulletin = Bulletin.objects.user_has_access(request.user).get(
            slug=slug
        )
        context = {"bulletin": bulletin, "slug": slug}

        return render(
            request=request,
            template_name="aa_bulletin_board/bulletin.html",
            context=context,
        )
    except Bulletin.DoesNotExist:
        messages.warning(
            request=request,
            message=_(
                "The bulletin you are looking for does either not exist, or you don't "
                "have access to it."
            ),
        )

        return redirect(to="aa_bulletin_board:dashboard")


@login_required
@permission_required(perm="aa_bulletin_board.manage_bulletins")
def edit_bulletin(request: WSGIRequest, slug: str) -> HttpResponse:
    """
    Edit an existing bulletin

    :param request:
    :type request:
    :param slug:
    :type slug:
    :return:
    :rtype:
    """

    try:
        bulletin: Bulletin = Bulletin.objects.get(slug=slug)
    except Bulletin.DoesNotExist:
        messages.warning(
            request=request,
            message=_("The bulletin you are trying to edit does not exist."),
        )

        return redirect(to="aa_bulletin_board:dashboard")

    # Create the form instance
    form = BulletinForm(instance=bulletin)

    # If we have a POST request
    if request.method == "POST":
        # Create the form instance and populate it with the data from te POST request
        form = BulletinForm(data=request.POST, instance=bulletin)

        # Check whether it's valid
        if form.is_valid():
            bulletin__title = form.cleaned_data["title"]
            bulletin__content = form.cleaned_data["content"]
            bulletin__updated_date = timezone.now()

            # Save the updated bulletin
            bulletin.title = bulletin__title
            bulletin.content = bulletin__content
            bulletin.updated_date = bulletin__updated_date
            bulletin.groups.set(form.cleaned_data["groups"])
            bulletin.save()

            messages.success(
                request=request, message=_(f'Bulletin "{bulletin__title}" updated.')
            )

            return redirect(to="aa_bulletin_board:view_bulletin", slug=bulletin.slug)

    context = {"form": form, "existing_bulletin": True, "bulletin": bulletin}

    return render(
        request=request,
        template_name="aa_bulletin_board/edit-bulletin.html",
        context=context,
    )


@login_required
@permission_required("aa_bulletin_board.manage_bulletins")
def remove_bulletin(request: WSGIRequest, slug: str) -> HttpResponseRedirect:
    """
    Remove bulletin

    :param request:
    :type request:
    :param slug:
    :type slug:
    :return:
    :rtype:
    """

    try:
        bulletin: Bulletin = Bulletin.objects.get(slug=slug)

        messages.success(
            request=request, message=_(f'Bulletin "{bulletin.title}" deleted.')
        )

        bulletin.delete()
    except Bulletin.DoesNotExist:
        messages.warning(
            request=request,
            message=_("The bulletin you are trying to delete does not exist."),
        )

    return redirect(to="aa_bulletin_board:dashboard")



# Generated by Django 4.0.10 on 2023-06-23 14:38


# Django
from django.conf import settings
from django.db import migrations, models

# AA Bulletin Board
import aa_bulletin_board.models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("auth", "0012_alter_user_first_name_max_length"),
        ("aa_bulletin_board", "0003_group_restrictions"),
    ]

    operations = [
        migrations.AlterField(
            model_name="bulletin",
            name="content",
            field=models.TextField(blank=True, verbose_name="Content"),
        ),
        migrations.AlterField(
            model_name="bulletin",
            name="created_by",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=models.SET(aa_bulletin_board.models.get_sentinel_user),
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="User",
            ),
        ),
        migrations.AlterField(
            model_name="bulletin",
            name="created_date",
            field=models.DateTimeField(
                auto_now_add=True, db_index=True, verbose_name="Created"
            ),
        ),
        migrations.AlterField(
            model_name="bulletin",
            name="groups",
            field=models.ManyToManyField(
                blank=True,
                related_name="aa_bulletin_board_group_restriction",
                to="auth.group",
                verbose_name="Group restrictions",
            ),
        ),
        migrations.AlterField(
            model_name="bulletin",
            name="title",
            field=models.CharField(max_length=255, verbose_name="Title"),
        ),
        migrations.AlterField(
            model_name="bulletin",
            name="updated_date",
            field=models.DateTimeField(
                auto_now=True, null=True, verbose_name="Updated"
            ),
        ),
    ]



"""
Pages url config
"""

# Django
from django.urls import path

# AA Bulletin Board
from aa_bulletin_board import views

app_name: str = "aa_bulletin_board"

urlpatterns = [
    path(route="", view=views.dashboard, name="dashboard"),
    path(route="create/", view=views.create_bulletin, name="create_bulletin"),
    path(route="bulletin/<slug:slug>/", view=views.view_bulletin, name="view_bulletin"),
    path(route="edit/<slug:slug>/", view=views.edit_bulletin, name="edit_bulletin"),
    path(
        route="remove/<slug:slug>/", view=views.remove_bulletin, name="remove_bulletin"
    ),
]



"""
Test bulletins
"""

# Third Party
from faker import Faker

# Django
from django.contrib.auth.models import Group, User
from django.contrib.messages import get_messages
from django.test import TestCase
from django.urls import reverse

# AA Bulletin Board
from aa_bulletin_board.forms import BulletinForm
from aa_bulletin_board.helper.string import string_cleanup
from aa_bulletin_board.models import Bulletin, get_sentinel_user
from aa_bulletin_board.tests.utils import create_fake_user

fake = Faker()


class TestGetSentinelUser(TestCase):
    """
    Tests for the sentinel user
    """

    def test_should_create_user_when_it_does_not_exist(self) -> None:
        """
        Test should create a sentinel user when it doesn't exist

        :return:
        :rtype:
        """

        # when
        user = get_sentinel_user()

        # then
        self.assertEqual(first=user.username, second="deleted")

    def test_should_return_user_when_it_does(self) -> None:
        """
        Test should return sentinel user when it exists

        :return:
        :rtype:
        """

        # given
        User.objects.create_user(username="deleted")

        # when
        user = get_sentinel_user()

        # then
        self.assertEqual(first=user.username, second="deleted")


class TestBulletins(TestCase):
    """
    Test Bulletins
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")

        # User cannot access bulletins
        cls.user_1001 = create_fake_user(
            character_id=1001, character_name="Peter Parker"
        )

        # User can access bulletins
        cls.user_1002 = create_fake_user(
            character_id=1002,
            character_name="Bruce Wayne",
            permissions=["aa_bulletin_board.basic_access"],
        )

        # User can manage bulletins
        cls.user_1003 = create_fake_user(
            character_id=1003,
            character_name="Clark Kent",
            permissions=[
                "aa_bulletin_board.basic_access",
                "aa_bulletin_board.manage_bulletins",
            ],
        )

    def test_should_remove_bulletin(self) -> None:
        """
        Test if a bulletin should be removed

        :return:
        :rtype:
        """

        # given
        bulletin_title = fake.sentence()
        bulletin = Bulletin.objects.create(
            title=bulletin_title,
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        self.client.force_login(user=self.user_1003)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_bulletin_board:remove_bulletin", args=[bulletin.slug]
            )
        )
        messages = list(get_messages(request=response.wsgi_request))

        # then
        self.assertRedirects(response=response, expected_url="/bulletin-board/")
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]), second=f'Bulletin "{bulletin_title}" deleted.'
        )

    def test_should_raise_does_not_exist_exception_when_delete_bulletin_not_found(
        self,
    ) -> None:
        """
        Test if a bulletin that doesn't exist should be removed

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title=fake.sentence(),
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        self.client.force_login(user=self.user_1003)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_bulletin_board:remove_bulletin", args=["foobarsson"]
            )
        )
        messages = list(get_messages(response.wsgi_request))

        self.assertRaises(expected_exception=bulletin.DoesNotExist)
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second="The bulletin you are trying to delete does not exist.",
        )

    def test_should_raise_does_not_exist_exception_when_edit_bulletin_not_found(
        self,
    ) -> None:
        """
        Test if a bulletin that doesn't exist should be edited

        :return:
        :rtype:
        """

        # given
        bulletin = Bulletin.objects.create(
            title=fake.sentence(),
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )
        self.client.force_login(user=self.user_1003)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_bulletin_board:edit_bulletin", args=["foobarsson"]
            )
        )
        messages = list(get_messages(request=response.wsgi_request))

        self.assertRaises(expected_exception=bulletin.DoesNotExist)
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second="The bulletin you are trying to edit does not exist.",
        )

    def test_should_translate_russian_letters_in_slug(self) -> None:
        """
        Test that russian letters in a slug are translated

        :return:
        :rtype:
        """

        bulletin = Bulletin.objects.create(
            title="дрифтерке, рорки в док",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        self.assertEqual(first=bulletin.slug, second="drifterke-rorki-v-dok")

    def test_should_return_cleaned_message_string_on_bulletin_creation(self) -> None:
        """
        Test should return a clean/sanitized message string
        when a new bulletin is created

        :return:
        :rtype:
        """

        # given
        dirty_message = (
            'this is a script test. <script type="text/javascript">alert('
            "'test')</script>and this is style test. <style>.MathJax, "
            ".MathJax_Message, .MathJax_Preview{display: none}</style>end tests."
        )
        cleaned_message = string_cleanup(string=dirty_message)
        bulletin = Bulletin.objects.create(
            title="Foobar",
            content=dirty_message,
            created_by=self.user_1001,
        )

        self.assertEqual(first=bulletin.content, second=cleaned_message)

    def test_bulletin_slug_creation(self) -> None:
        """
        Test slug creation

        :return:
        :rtype:
        """

        bulletin = Bulletin.objects.create(
            title="This is a bulletin",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        bulletin_2 = Bulletin.objects.create(
            title="This is a bulletin",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        bulletin_3 = Bulletin.objects.create(
            title="This is a bulletin",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        self.assertEqual(first=bulletin.slug, second="this-is-a-bulletin")
        self.assertEqual(first=bulletin_2.slug, second="this-is-a-bulletin-1")
        self.assertEqual(first=bulletin_3.slug, second="this-is-a-bulletin-2")

    def test_should_return_bulletin_title_as_model_object_string_name(self) -> None:
        """
        Test should return the object's string name

        :return:
        :rtype:
        """

        bulletin = Bulletin.objects.create(
            title="This is a bulletin",
            content=f"<p>{fake.sentence()}</p>",
            created_by=self.user_1001,
        )

        self.assertEqual(first=str(bulletin), second="This is a bulletin")

    def test_form_clean_content_method(self) -> None:
        """
        Test the clean_ method of the form

        :return:
        :rtype:
        """

        dirty_message = (
            'this is a script test. <script type="text/javascript">alert('
            "'test')</script>and this is style test. <style>.MathJax, "
            ".MathJax_Message, .MathJax_Preview{display: none}</style>end tests."
        )
        cleaned_message = string_cleanup(string=dirty_message)
        data = {"title": "This is a title", "content": dirty_message}

        form = BulletinForm(data=data)

        self.assertTrue(expr=form.is_valid())
        self.assertEqual(first=form.cleaned_data["content"], second=cleaned_message)
